/*	
*      CHPOX (CHeckPOinting for linuX)
*	
*      Copyright (C) 2002, Olexander O. Sudakov <csaa@mail.univ.kiev.ua>
*      Copyright (C) 2002, Eugeniy S. Meshcheryakov <eugen@univ.kiev.ua>
* 
*      This program is free software; you can redistribute it and/or modify
*      it under the terms of the GNU General Public License as published by
*      the Free Software Foundation; either version 2 of the License, or
*      (at your option) any later version.
* 
*      This program is distributed in the hope that it will be useful,
*      but WITHOUT ANY WARRANTY; without even the implied warranty of
*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*      GNU General Public License for more details.
* 
*      You should have received a copy of the GNU General Public License
*      along with this program; if not, write to the Free Software
*      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#define __KERNEL_SYSCALLS__

#include <linux/config.h>
#if defined(CONFIG_SMP) && ! defined(__SMP__)
#define __SMP__
#endif
#if defined(CONFIG_MODVERSIONS) && ! defined(MODVERSIONS)
#define MODVERSIONS
#endif
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/errno.h>
#include <linux/smp_lock.h>
#include <linux/unistd.h>
#include <linux/personality.h>
#include <asm/pgtable.h>
#include <asm/processor.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <linux/proc_fs.h>
#include <linux/signal.h>
#include <linux/spinlock.h>
#include <linux/rwsem.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <linux/highmem.h>
#include <linux/limits.h>
#include <linux/types.h>
#include <linux/unistd.h>

#include "vmadump.h"
#include "chpox.h"
#include "chpox_config.h"

#ifdef CHPOX_OPEN_MOSIX
#include <linux/hpc.h>
#elif  CHPOX_MOSIX
#include <linux/mosix.h>
#endif

#ifdef MODULE
int errno;
#endif

static char vmad_magic[3] = VMAD_MAGIC;

static struct proc_dir_entry *dump_dir;
static struct proc_dir_entry *dump_it;
static struct proc_dir_entry *dump_conf;
static struct proc_dir_entry *dump_info;
/*

*/
static const char *ff = "/tmp";

static LIST_HEAD(ckpt_list);
static DECLARE_RWSEM(ckpt_lock);
#if 0
/* XXX may not work */
static LIST_HEAD(ckpt_pipes);
static DECLARE_MUTEX(pipes_lock);
static LIST_HEAD(ckpt_restored_pipes);
static DECLARE_MUTEX(restored_pipes_lock);
#endif
MODULE_LICENSE("GPL");

static inline _syscall0(pid_t, fork);
static inline _syscall1(int, chdir, const char *, path);
static inline _syscall2(int, dup2, int, oldfd, int, newfd);
#if 0
static inline _syscall1(int, pipe, int *, filedes);
#endif
static int create_chpox_entry(struct ckpt_data *pdt, struct task_struct *o,
		       int interval, int signal);

/*--------------------------------------------------------------------
 *  Some utliity stuff for reading and writing and misc kernel syscalls 
 (from vmadump).
*------------------------------------------------------------------*/

static ssize_t default_read_file(struct vmadump_map_ctx *ctx,
				 struct file *file, void *buf, size_t count)
{
	ssize_t ret = -EBADF;
	if (file->f_mode & FMODE_READ) {
		ret = locks_verify_area(FLOCK_VERIFY_READ,
					file->f_dentry->d_inode, file,
					file->f_pos, count);
		if (!ret) {
			ssize_t(*read) (struct file *, char *, size_t,
					loff_t *);
			ret = -EINVAL;
			if (file->f_op && (read = file->f_op->read) != NULL)
				ret = read(file, buf, count, &file->f_pos);
		}
	}
	return ret;
}

static ssize_t read_user(struct vmadump_map_ctx *ctx,
			 struct file *file, void *buf, size_t count)
{
	ssize_t r, bytes = count;
	ssize_t(*rfunc) (struct vmadump_map_ctx * ctx, struct file * file,
			 void *buf, size_t count);
	rfunc = (ctx && ctx->read_file) ? ctx->read_file : default_read_file;
	while (bytes) {
		r = rfunc(ctx, file, buf, bytes);
		if (r < 0)
			return r;
		if (r == 0)
			return count - bytes;
		bytes -= r;
		buf += r;
	}
	return count;
}

static ssize_t read_kern(struct vmadump_map_ctx *ctx,
			 struct file *file, void *buf, size_t count)
{
	ssize_t err;
	mm_segment_t oldfs;
	oldfs = get_fs();
	set_fs(KERNEL_DS);
	err = read_user(ctx, file, buf, count);
	set_fs(oldfs);
	return err;
}

static ssize_t default_write_file(struct vmadump_map_ctx *ctx,
				  struct file *file, const void *buf,
				  size_t count)
{
	ssize_t ret = -EBADF;
	if (file->f_mode & FMODE_WRITE) {
		struct inode *inode = file->f_dentry->d_inode;
		ret = locks_verify_area(FLOCK_VERIFY_WRITE, inode, file,
					file->f_pos, count);
		if (!ret) {
			ssize_t(*write) (struct file *, const char *, size_t,
					 loff_t *);
			ret = -EINVAL;
			if (file->f_op && (write = file->f_op->write) != NULL)
				ret = write(file, buf, count, &file->f_pos);
		}
	}
	return ret;
}

static ssize_t write_user(struct vmadump_map_ctx *ctx, struct file *file,
			  const void *buf, size_t count)
{
	ssize_t w, bytes = count;
	ssize_t(*wfunc) (struct vmadump_map_ctx * ctx, struct file * file,
			 const void *buf, size_t count);
	wfunc = (ctx
		 && ctx->write_file) ? ctx->write_file : default_write_file;
	while (bytes) {
		w = wfunc(ctx, file, buf, bytes);
		if (w < 0)
			return w;
		if (w == 0)
			return count - bytes;
		bytes -= w;
		buf += w;
	}
	return count;
}

static ssize_t write_kern(struct vmadump_map_ctx *ctx, struct file *file,
			  const void *buf, size_t count)
{
	int err;
	mm_segment_t oldfs;
	oldfs = get_fs();
	set_fs(KERNEL_DS);
	err = write_user(ctx, file, buf, count);
	set_fs(oldfs);
	return err;
}

extern void *sys_call_table[];

/*
  End from vmadump
*/

#if 0
static int restore_pipe_file(struct open_files *ofs)
{
	struct list_head *i = &ckpt_restored_pipes;
	struct ckpt_pipe_restore_entry *ent = NULL;
	struct inode *pipe_ino = NULL;
	
	down(&restored_pipes_lock);

	list_for_each(i, &ckpt_restored_pipes) {
		ent = list_entry(i, struct ckpt_pipe_restore_entry, list);
		if (ent->ino == ofs->u.pipes.inode)
			break;
	}

	if (i == &ckpt_restored_pipes) {
		/* not found */
		mm_segment_t oldfs;
		int pipe_des[2];
		struct file *r_file, *w_file;
		struct file *dump;
		static char filename[PATH_MAX];
		
		ent = kmalloc(sizeof(*ent), GFP_KERNEL);
		if (!ent) {
			printk(KERN_ERR "chpox: out of memory\n");
			goto out_up;
		}
		
		oldfs = get_fs(); set_fs(KERNEL_DS);
		if (pipe(pipe_des) == -1) {
			printk(KERN_ERR "chpox cannot create pipe (%d)\n", errno);
			set_fs(oldfs);
			kfree(ent);
			goto out_up;
		}
		set_fs(oldfs);

		r_file = fcheck(pipe_des[0]); get_file(r_file);
		w_file = fcheck(pipe_des[1]); get_file(w_file);
		ent->r_file = r_file;
		ent->w_file = w_file;
		ent->ino = ofs->u.pipes.inode;
		
		snprintf(filename, PATH_MAX, "%s/%s+inode-%lu", ff, DUMP_PROC_DIR, ent->ino);
		dump = filp_open(filename, O_RDONLY | O_NOFOLLOW, 0660);
		if (IS_ERR(dump))
			printk(KERN_ERR "chpox: cannot open file %s\n", filename);
		else {
			unsigned long nrefs;
			unsigned long size;
			int fd;
			
			pipe_ino = r_file->f_dentry->d_inode;
			read_kern(NULL, dump, &nrefs, sizeof(nrefs));
			read_kern(NULL, dump, &size, sizeof(size));
			ent->nrefs = nrefs;
			
			down(PIPE_SEM(*pipe_ino));
		
			/* FIXME: check size of data */
			read_kern(NULL, dump, PIPE_BASE(*pipe_ino), size);
			PIPE_LEN(*pipe_ino) = size;
			PIPE_START(*pipe_ino) = 0;
			list_add(&ent->list, &ckpt_restored_pipes);
			if (ofs->u.pipes.rdwr) {
				/* open for writing */
				close(pipe_des[0]);
				fd = pipe_des[1];
			}
			else {
				close(pipe_des[1]);
				fd = pipe_des[0];
			}
			if (ofs->fd != fd) {
				dup2(fd, ofs->fd);
				close(fd);
			}
			ent->nrefs--;
			filp_close(dump, 0);
		}
	}
	else {
		/* found */
		int fd;
		
		pipe_ino = ent->r_file->f_dentry->d_inode;
		
		fd = get_unused_fd();
		if (ofs->u.pipes.rdwr) {
			/* for writing */
			get_file(ent->w_file);
			fd_install(fd, ent->w_file);
			PIPE_WRITERS(*pipe_ino)++;	
		}
		else {
			get_file(ent->r_file);
			fd_install(fd, ent->r_file);
			PIPE_READERS(*pipe_ino)++;
		}
		ent->nrefs--;
		if (fd != ofs->fd) {
			dup2(fd, ofs->fd);
			close(fd);
		}
	}
out_up:
	if (!ent->nrefs) {
		fput(ent->r_file);
		fput(ent->w_file);
		list_del(&ent->list);
		kfree(ent);
		wake_up_interruptible(PIPE_WAIT(*pipe_ino));
		up(PIPE_SEM(*pipe_ino));
	}
	up(&restored_pipes_lock);
	return 0;
}
#endif

/*
  restores opened files (simalar to CRACK)
*/
static int restore_files(struct vmadump_map_ctx *ctx, struct file *file)
{
	int filecnt, r = 0, ret = 0;
	char *buf = 0, i;
	struct file *nfile;
	mm_segment_t oldfs;

	/* Restore CWD */
	{
		int size;
		read_kern(ctx, file, &size, sizeof(size));
		buf = (char *)kmalloc(size + 1, GFP_KERNEL);
		read_kern(ctx, file, buf, size);
		buf[size] = '\0';
		oldfs = get_fs();
		set_fs(KERNEL_DS);
		if (chdir(buf) ==  -1)
			printk(KERN_ERR "chpox: cannot set cwd (%d)\n", errno);;
		set_fs(oldfs);
		kfree(buf);

	}

	r = read_kern(ctx, file, &filecnt, sizeof(filecnt));
	if (r != sizeof(filecnt))
		goto out;
	ret += r;
	for (i = 0; i < filecnt; i++) {
		//struct file *fdes = NULL;
		//struct inode *inode = NULL;
		struct open_files open_files;
		r = read_kern(ctx, file, &open_files,
			      sizeof(struct open_files));
		//printk("read %d\n",i);
		if (r != sizeof(struct open_files))
			goto out;
		ret += r;
		//fdes = current->files->fd[open_files.fd];
		//inode = fdes->f_dentry->d_inode;
		switch (open_files.type) {
		case CKPT_DUP:{
				if (dup2(open_files.fd, open_files.u.dup.dupfd) == -1) {
					printk(KERN_ERR "chpox: restore_files: dup %d failed\n", open_files.fd);
					goto out;
				}
				break;
			}
		case CKPT_FILE:{
				char *line = kmalloc(open_files.entry_size + 1,
						GFP_KERNEL);
				int fd1;
				if (!line) {
					printk(KERN_ERR "chpox: restore_files: out of memory\n");
					goto out;
				}
				r = read_kern(ctx, file, line,
					      open_files.entry_size);
				line[open_files.entry_size] = '\0';
				if (OPEN_SAFE(open_files.fd, line)) {
					oldfs = get_fs();
					set_fs(KERNEL_DS);
					fd1 = open(line, open_files.u.file.flags,
						       open_files.u.file.
						       mode);
					set_fs(oldfs);
					if (fd1 == -1) {
						printk(KERN_ERR "chpox: cannot open file %s\n", line);
						if (line)
							kfree(line);
						line = NULL;
						goto out;
					}
					nfile = fcheck_files(current->files,
							     fd1);
					if (!nfile) {
						if (line)
							kfree(line);
						line = NULL;
						printk(KERN_ERR "chpox: fcheck_files no file\n");
						goto out;
					}
					nfile->f_pos = open_files.u.file.file_pos;
					if (fd1 != open_files.fd) {
						if (dup2(fd1, open_files.fd) == -1) {
							printk(KERN_ERR "chpox: cannot dup2 file %d\n", open_files.fd);
						}
						close(fd1);
					}
				}
				if (line)
					kfree(line);
				line = NULL;
				break;
			}
		case CKPT_PIPE:{
#if 0
					if (restore_pipe_file(&open_files))
						goto out;
#else
					printk(KERN_ERR "chpox: pipes not supported\n");
					goto out;
#endif
			}
			break;
		case CKPT_SOCK:
			printk(KERN_ERR "chpox: socket not supported\n");
			goto out;
		}
	}
      
out:
	{
		int flag;
		read_kern(ctx, file, &flag, sizeof(flag));
		if (flag && current->tty && current->tty->termios) {
			read_kern(ctx, file, current->tty->termios,
				  sizeof(struct termios));
		} else 
			file->f_pos += sizeof(struct termios);
	}

	return ret;

}

static int restore_children(struct vmadump_map_ctx *ctx, struct file *file,
			    struct pt_regs *regs)
{
	int chld_count;
	int fname_len;
	int i;
	int r;
	int ch_pid;
	mm_segment_t oldfs;
	static char *envp[] = { "HOME=/",
				"TERM=linux",
				"PATH=/bin:/sbin:/usr/bin:/usr/sbin",
				NULL
	};
	/* XXX May not work on SMP (static variables) XXX */
	static char *argv[] = { NULL, NULL };
	static char f_name[1024];

	r = read_kern(ctx, file, &chld_count, sizeof(chld_count));
	if (r < 0) {
		return -1;
	}
	printk(KERN_DEBUG "chpox: pid %d chld_count %d\n", current->pid, chld_count);
	for (i = 0; i < chld_count; i++) {
		r = read_kern(ctx, file, &fname_len, sizeof(fname_len));
		if (r < 0) {
			return -1;
		}
		if (fname_len > sizeof(f_name)) {
			printk(KERN_ERR "chpox: file name too long\n");
			return -1;
		}
		
		r = read_kern(ctx, file, f_name, fname_len);
		if (r < 0) 
			return -1;

		oldfs = get_fs();
		set_fs(KERNEL_DS);
		argv[0] = f_name;
		printk(KERN_INFO "chpox: executing '%s'\n", f_name);	
		switch (ch_pid = fork()) {
		case -1:
			printk(KERN_ERR "chpox: cannot clone (%d)\n", errno);
			break;
		case 0:
			/* child process */
			execve(f_name, argv, envp);
			_exit(1);
			break;
		default:
			printk(KERN_INFO "chpox: created new process %d\n", ch_pid);
			break;
		}
		set_fs(oldfs);
	}
	return 0;
}

#if 0
static int dump_pipe_file(struct file *file) {
	struct inode *pipe_ino = file->f_dentry->d_inode;
	struct list_head *i = &ckpt_pipes;
	struct ckpt_pipe_entry *ent = NULL;	
	
	printk("chpox: in dump_pipe_file, inode %lu\n", pipe_ino->i_ino);
	down(&pipes_lock);
	list_for_each(i, &ckpt_pipes) {
		ent = list_entry(i, struct ckpt_pipe_entry, list);
		if (ent->inode == pipe_ino)
			break;
	}

	if (i == &ckpt_pipes) {
		/* entry not found */
		ent = kmalloc(sizeof(*ent), GFP_KERNEL);
		if (!ent) {
			printk(KERN_ERR "chpox: out of memory in dump_pipe_file\n");
			goto out_up;
		}
		ent->inode = pipe_ino;
		down(PIPE_SEM(*pipe_ino));
		ent->nrefs_orig = PIPE_READERS(*pipe_ino) + PIPE_WRITERS(*pipe_ino);
		ent->nrefs = ent->nrefs_orig - 1;
		list_add(&ent->list, &ckpt_pipes);
	}
	else {
		/* entry found */
		ent->nrefs--;
	}
	
	if (!ent->nrefs) {
		/* dump this inode to file */
		struct file *dump;
		static char filename[PATH_MAX];

		snprintf(filename, PATH_MAX, "%s/%s+inode-%lu", ff, DUMP_PROC_DIR, pipe_ino->i_ino);
		dump = filp_open(filename, O_WRONLY|O_CREAT|O_NOFOLLOW, 0660);
		if (IS_ERR(dump)) {
			printk(KERN_ERR "chpox: cannot open file %s\n", filename);
		}
		else {
			/* TODO */
			unsigned long size;
			ssize_t count, pipeofft;
			char *pipedata;
			
			printk(KERN_INFO "chpox: saving %s\n", filename);
			/* save # of refs first...*/
			write_kern(NULL, dump, &ent->nrefs_orig, sizeof(ent->nrefs_orig));
			/* ...amount of data in pipe ...*/
			size = PIPE_LEN(*pipe_ino);
			write_kern(NULL, dump, &size, sizeof(size));
			/* ... and content of this pipe */
			count = PIPE_LEN(*pipe_ino);
			pipeofft = PIPE_START(*pipe_ino);
			
			while (count > 0) {
				ssize_t chars = PIPE_MAX_RCHUNK(*pipe_ino);
				pipedata = PIPE_BASE(*pipe_ino) + pipeofft;
				if (chars > count)
					chars = count;
				write_kern(NULL, dump, pipedata, chars);
				pipeofft += chars;
				pipeofft &= PIPE_SIZE-1;
				count -= chars;
			}

			filp_close(dump, 0);
		}
		/* free inode and entry */
		wake_up_interruptible(PIPE_WAIT(*pipe_ino));
		up(PIPE_SEM(*pipe_ino));
		list_del(&ent->list);
		kfree(ent);
	}
out_up:
	up(&pipes_lock);
	return 0;
}
#endif 

static int dump_files(struct vmadump_map_ctx *ctx, struct file *file)
{
	struct files_struct *files;
	//struct file *cur_f;
	//int dump_fd;
	int filecnt = 0, fds = 0, j = 0, r = 0, ret = 0;
	char *buffer = 0;
	char *line = 0;
	struct ckpt_fdcache_struct *fdcache = 0;
	struct open_files open_files;
	//struct open_files_hdr open_files_hdr;
	int fdcache_size = 0;
	//int dump_fd;

	if (!buffer)
		buffer = (char *)__get_free_page(GFP_KERNEL);
	if (!buffer)
		goto out;

	/* Store CWD */
	{
		int size;
		struct dentry *dentry = dget(current->fs->pwd);
		struct vfsmount *mnt = mntget(current->fs->pwdmnt);
		line = d_path(dentry, mnt, buffer, PAGE_SIZE);
		buffer[PAGE_SIZE - 1] = 0;
		size = buffer + PAGE_SIZE - line;
		dput(dentry);
		mntput(mnt);
		r = write_kern(ctx, file, &size, sizeof(size));
		if (r != sizeof(size))
			goto out;
		ret += r;
		r = write_kern(ctx, file, line, size);
		if (r != size)
			goto out;
		ret += r;
	}
	// how many files are opened (filecnt)
	files = current->files;
	for (;;) {
		unsigned long set;
		fds = j * __NFDBITS;
		if (fds >= files->max_fdset || fds >= files->max_fds)
			break;
		set = files->open_fds->fds_bits[j++];
		while (set) {
			if (set & 1) {
				if (ckpt_icheck_task(current, fds))
					filecnt++;
			}
			fds++;
			set >>= 1;
		}
	}
	r = write_kern(ctx, file, &filecnt, sizeof(filecnt));
	if (r != sizeof(filecnt))
		goto out;
	ret += r;

	fdcache =
		(struct ckpt_fdcache_struct *)
		kmalloc(sizeof(struct ckpt_fdcache_struct) * filecnt,
			GFP_KERNEL);
	if (!fdcache)
		goto out;


	// save files
	j = 0;
	for (;;) {
		unsigned long set;
		fds = j * __NFDBITS;
		if (fds >= files->max_fdset || fds >= files->max_fds)
			break;
		set = files->open_fds->fds_bits[j++];
		while (set) {
			if (set & 1) {
				struct file *fdes = fcheck_files(files, fds);
				struct dentry *dent =
					fdes ? fdes->f_dentry : NULL;
				struct inode *inode =
					dent ? dent->d_inode : NULL;
				int i;
				if (!inode)
					goto next;

				open_files.fd = fds;
				// check whether this inode has appeared before
				for (i = 0; i < fdcache_size; i++) {
					if (fdes == fdcache[i].file) {
						// cache hit
						open_files.type = CKPT_DUP;
						open_files.u.dup.dupfd =
							fdcache[i].fd;
						open_files.entry_size = 0;
						r = write_kern(ctx, file,
							       &open_files,
							       sizeof
							       (open_files));
						if (r != sizeof(open_files))
							goto out;
						ret += r;
						goto next;
					}
				}
				// if not a dup, push to the cache
				fdcache[fdcache_size].fd = fds;
				fdcache[fdcache_size].inode = inode; /* XXX May be not needed */
				fdcache[fdcache_size].file = fdes;
				fdcache_size++;

				if (!inode)
					printk(KERN_ERR "chpox: fd %d has no entry\n", fds);
				else if (S_ISSOCK(inode->i_mode)) {
					/* TODO Add support for sockets */
					printk(KERN_ERR "chpox: fd %d is socket - unsupported\n", fds);
					ret = -ENOSYS;
					goto out;
				}
				else if ((S_ISFIFO(inode->i_mode))) {	// unnamed pipe
#if 0
					open_files.type = CKPT_PIPE;
					open_files.fd = fds;
					open_files.u.pipes.inode = inode->i_ino;
					open_files.u.pipes.rdwr = fdes->f_flags & O_WRONLY;
					r = write_kern(ctx, file, &open_files, sizeof(open_files));
					if (r != sizeof(open_files)) 
						goto out;
					ret += r;
					if (dump_pipe_file(fdes))
						goto out;
#else
					printk(KERN_ERR "chpox: pipes not supported\n");
					ret = -ENOSYS;
					goto out;
#endif
				}
				else if (IS_FD_NAMED(fdes)) {	//regular file
					struct dentry *dentry = dget(dent);
					struct vfsmount *mnt =
						mntget(fdes->f_vfsmnt);
					line = d_path(dentry, mnt, buffer,
						      PAGE_SIZE);
					buffer[PAGE_SIZE - 1] = 0;
					dput(dentry);
					mntput(mnt);

					open_files.type = CKPT_FILE;
					open_files.fd = fds;
					open_files.u.file.file =
						(unsigned long)fdes;
					open_files.u.file.flags =
						fdes->f_flags;
					open_files.u.file.mode = fdes->f_mode;
					open_files.u.file.file_pos =
						fdes->f_pos;

					open_files.entry_size =
						buffer + PAGE_SIZE - line;

					r = write_kern(ctx, file, &open_files,
						       sizeof(open_files));
					if (r != sizeof(open_files))
						goto out;
					ret += r;
					r = write_kern(ctx, file, line,
						       open_files.entry_size);
					if (r != open_files.entry_size)
						goto out;
					ret += r;
				}
				else
					printk(KERN_ERR "chpox: unknown file type, cannot handle\n");
			}

		      next:
			fds++;
			set >>= 1;
		}
	}
	/* Store terminal state */
	{
		int flag = 0;
		if (current->tty && current->tty->termios) {
			flag = 1;
		}
		write_kern(ctx, file, &flag, sizeof(flag));
		if (flag) {
			write_kern(ctx, file, current->tty->termios,
				   sizeof(struct termios));
		}
	}

      out:

	if (buffer)
		free_page((unsigned long)buffer);
	if (fdcache)
		kfree(fdcache);

	return ret;
}

/*
 *
 * *
 */
static int unchpox_proc(struct vmadump_map_ctx *ctx, struct file *file,
		 struct pt_regs *regs)
{
	int r = 0;
#ifdef CONFIG_MOSIX
	/* deny migration during restart */
	int was_locked;
	if ((!current->mm) || (current->mosix.dflags & DDEPUTY)
	    || (current->mosix.dflags & DREMOTE)) {
		// It is better to get process  home ???
		return -1;
	}
	/* From mos/mig.c */
	task_lock(current);
	was_locked = ((current->mosix.stay & DNOMIGRATE) == DNOMIGRATE);
	current->mosix.stay |= DNOMIGRATE;
	task_unlock(current);

#else
	/* brobably should no happen ??? */
	if (!current->mm)
		return -1;
#endif
	r = vmadump_thaw_proc(ctx, file, regs);
#ifdef CONFIG_MOSIX
	/*chamge lock state back */
	if (!was_locked) {
		task_lock(current);
		current->mosix.stay &= ~DNOMIGRATE;
		task_unlock(current);
	}
#endif

	if (r < 0) {
		printk(KERN_ERR "chpox: vmadump_thaw_proc error\n");
		return -1;
	}
	r = restore_files(ctx, file);
	if (r < 0) {
		printk(KERN_ERR "chpox: restore_files error\n");
		return -1;
	}
	r = restore_children(ctx, file, regs);

	return 0;
}

/*
 *
 */
static int chpox_proc(struct vmadump_map_ctx *ctx, struct file *file,
	       struct pt_regs *regs, int flags)
{
	int r = 0;
	/* brobably should not happen ??? */
	if (!current->mm)
		return -1;
	r = vmadump_freeze_proc(ctx, file, regs, flags);
	if (r < 0) {
		printk(KERN_ERR "chpox: vmadump_freeze_proc error\n");
		return -1;
	}

	read_lock(&current->files->file_lock);
	r = dump_files(ctx, file);
	read_unlock(&current->files->file_lock);
	if (r < 0) {
		printk(KERN_ERR "chpox: dump_files error\n");
		return -1;
	}

	return 0;
}

/*

*/
static void tsk_block_all_signals(int (*notifier) (void *priv), void *priv,
			   struct task_struct *tsk)
{
	struct ckpt_data *dt = (struct ckpt_data *)priv;


	if (!priv)
		return;
	sigemptyset(&dt->mask);
	sigaddset(&dt->mask, dt->signal);

	spin_lock(&tsk->sigmask_lock);
	tsk->notifier_mask = &dt->mask;
	tsk->notifier_data = priv;
	tsk->notifier = notifier;
	spin_unlock(&tsk->sigmask_lock);
}

/*
*
*/
#ifdef __CHPOX_DEL_SIGNAL__
static int (*chpox_del_sig) (int sig, struct sigpending * list,
			     siginfo_t * info) = (void *)__CHPOX_DEL_SIGNAL__;
#else
#error "missing file chpox_config.h"
#endif
/*
  
*/
static int submit_children(struct ckpt_data *dt)
{
	struct task_struct *p;
	struct ckpt_data *dt1 = NULL;
	struct list_head *lh = NULL;
	int chld_count = 0;
	if (!dt)
		return -1;
	//currnet->state = TASK_UNINTERRUPTIBLE;
	read_lock(&tasklist_lock);
	for (p = current->p_cptr; p; p = p->p_osptr) {
		get_task_struct(p);
		read_unlock(&tasklist_lock);
		chld_count++;
		down_write(&ckpt_lock);
		list_for_each(lh, &ckpt_list) {
			dt1 = list_entry(lh, struct ckpt_data, list);
			if (dt1->pid == p->pid) {
				force_sig(dt1->signal, p);
				break;
			}
		}
		/* Child process not found in checkpoint list, so add it and send signal */
		if (lh == &ckpt_list)
			if (create_chpox_entry(dt, p, dt->interval, dt->signal) != -1)
				force_sig(dt->signal, p);
		up_write(&ckpt_lock);
		read_lock(&tasklist_lock);
		free_task_struct(p);
	}
	read_unlock(&tasklist_lock);
	return chld_count;
}

/*
 *
 * */
static int dump_chld_info(struct vmadump_map_ctx *ctx, int chld_count,
		   struct ckpt_data *dt)
{
	int ret = -1;
	struct ckpt_data *dt1 = NULL;
	struct list_head *lh = NULL;
	
	if (write_kern(ctx, dt->file, &chld_count, sizeof(chld_count)) != sizeof(chld_count)) {
		goto fail;
	}
	down_read(&ckpt_lock);
	list_for_each(lh, &ckpt_list) {
		dt1 = list_entry(lh, struct ckpt_data, list);
		if (dt1->parent == dt) {
			if (write_kern(ctx, dt->file, &dt1->fname_size,
				   sizeof(dt1->fname_size)) != sizeof(dt1->fname_size)) 
				goto fail_up;
			if (write_kern(ctx, dt->file, dt1->file_name,
				   dt->fname_size) != dt->fname_size)
				goto fail_up;
		}
	}
	ret = 0;
fail_up:
	up_write(&ckpt_lock);
fail:
	return ret;
}

#ifdef CONFIG_MOSIX
static int (*mosix_go)(int) = (int (*)(int))(__CHPOX_GO_HOME__);
#endif
/*
*
* */
static int ckpt_notifier(void *arg)
{
	int ret = 0;
	struct ckpt_data *dt = (struct ckpt_data *)arg;
	struct pt_regs *regs;
	struct inode *inode;
	struct siginfo info;
	int vmad_flags = 0;
	int chld_count = 0;;
	if (!arg)
		return 0;
	/* if do not like to checkpoint with trapped signal uncomment line below */
	//if(current->sig->action[dt->signal-1].sa.sa_handler != SIG_DFL) return 1;

	dt->count++;
	/* Drop signal lock */
	spin_unlock(&current->sigmask_lock);
	/* Get dt lock */
	//down_write(&dt->lock);

	//task_lock(current);
#ifdef CONFIG_MOSIX
/*Migrated processes are not allowed to be checkpointed. Try send them home*/
	/* do not checkpoint remote part of process */
	if (current->mosix.dflags & DREMOTE) {
		dt->flags = CKPT_FAILED;
		goto out;
	}

	if ((!current->mm) || (current->mosix.dflags & DDEPUTY)) {
		/*try to move process back home and return */
#if 0
		struct file *proc_goto;
		char txt[] = "0\n";
		//task_unlock(current); 
		proc_goto =
			filp_open("/proc/self/goto", O_WRONLY | O_NOFOLLOW,
				  0755);
		if (proc_goto) {
			write_kern(0, proc_goto, txt, 2);
			filp_close(proc_goto, 0);
		}
#endif
		(*mosix_go)(0);
		goto out;
	}
#else
	/* brobably should not happen ??? */
	if (!current->mm)
		goto out;
#endif
	(*chpox_del_sig) (dt->signal, &current->pending, &info);
	dt->nt++;

	dt->file =
		filp_open(dt->file_name, O_WRONLY | O_CREAT | O_NOFOLLOW,
			  0750);
	if (IS_ERR(dt->file)) {
		printk(KERN_ERR "chpox: can't open checkpoint file %s for %d\n",
		       dt->file_name, dt->pid);
		goto out;
	}
	printk(KERN_INFO "chpox: saving file %s\n", dt->file_name);

	/* security fs/exec.c:do_coredump */
	inode = dt->file->f_dentry->d_inode;
	if (inode->i_nlink > 1)
		goto close_fail;	/* multiple links - don't dump */
	if (d_unhashed(dt->file->f_dentry))
		goto close_fail;
	if (!S_ISREG(inode->i_mode))
		goto close_fail;
	if (!dt->file->f_op)
		goto close_fail;
	if (!dt->file->f_op->write)
		goto close_fail;


	regs = (((struct pt_regs *)(THREAD_SIZE + (unsigned long)current)) -
		1);


	if ((dt->interval & 1 << 1) == 1 << 1)
		vmad_flags |= VMAD_DUMP_EXEC;
	if ((dt->interval & 1 << 2) == 1 << 2)
		vmad_flags |= VMAD_DUMP_LIBS;

	if (chpox_proc(0, dt->file, regs, vmad_flags) >= 0) {
		dt->flags = CKPT_DONE;
		//get_fast_time(&dt->tv);   
		do_gettimeofday(&dt->tv);
	}
	else {
		dt->flags = CKPT_FAILED;
		goto close_fail;
	}
	if ((dt->interval & 1 << 3) == 1 << 3) {
		// submit children
		chld_count = submit_children(dt);
		if (dump_chld_info(0, chld_count, dt) < 0) {
			dt->flags = CKPT_FAILED;
			goto close_fail;
		}
	}
	else {
		int count = 0;
		/* Child count */
		if (write_kern(0, dt->file, &count, sizeof(count)) !=sizeof(count)) {
			dt->flags = CKPT_FAILED;
			goto close_fail;
		}
	}
      close_fail:
	filp_close(dt->file, NULL);
	/* Drop dt lock */
	//up_write(&dt->lock);

	/* Get lock again */
	spin_lock(&current->sigmask_lock);

	dt->count--;

      out:

#ifdef CONFIG_MOSIX
	/* remove lock state */
	current->mosix.stay &= ~DNOMIGRATE;
#endif
	if (dt->flags == CKPT_FAILED)
		printk(KERN_ERR "chpox: checkpoint failed\n");

	return ret;
}

static int create_chpox_entry(struct ckpt_data *pdt, struct task_struct *o,
		       int interval, int signal)
{
	int pn;
	char tmp_file_name[PATH_MAX + 1];
	char tmp_comm[17];
	struct ckpt_data *dt = NULL;
	struct timeval tv;
	/*malloc entry */
	dt = kmalloc(sizeof(struct ckpt_data), GFP_KERNEL);
	if (!dt) {
		printk(KERN_ERR "chpox: out of memory: %d\n", dt->pid);
		return -1;
	}
	/* construct file-name */
	do_gettimeofday(&tv);
	snprintf(tmp_comm, 16, "%s", o->comm);
	pn = snprintf(tmp_file_name, PATH_MAX, "%s/%s-%s-%05u.%lu.%lu", ff,
		      DUMP_PROC_DIR, tmp_comm, o->pid, tv.tv_sec, tv.tv_usec);
	if (pn < 1) {
		printk(KERN_ERR "chpox: can't create file name: %d\n", dt->pid);
		goto out;
	}
	dt->file_name = kmalloc(pn + 1, GFP_KERNEL);
	if (!dt->file_name) {
		printk(KERN_ERR "chpox: out of memory");
		goto out;
	}
	dt->fname_size = pn + 1;
	strcpy(dt->file_name, tmp_file_name);
	dt->pid = o->pid;
	dt->interval = interval;
	dt->flags = CKPT_CREATED;
	dt->signal = signal;
	dt->nt = 0;
	dt->tv.tv_sec = 0;
	dt->tv.tv_usec = 0;
	dt->count = 0;
	dt->parent = pdt;
	MOD_INC_USE_COUNT;
	list_add(&dt->list, &ckpt_list);
	/* set notifier */
	tsk_block_all_signals(ckpt_notifier, dt, o);
	return 0;
      out:
	kfree(dt);
	return -1;

}

/*

*/
static int ckpt_write(struct file *f, const char *buf, unsigned long len, void *off)
{
	char ll[20];
	int pid = 0;
	int interval = 0;
	int i = 0;
	int ret = len;
	struct task_struct *o = NULL;
	struct ckpt_data *dt = NULL;
	struct list_head *lh = NULL;
	
	int signal;

	MOD_INC_USE_COUNT;
	for (i = 0; i < len && i < 20; i++)
		get_user(ll[i], buf + i);
	ll[i] = 0;
	if (-1 == sscanf(ll, "%d:%d:%d", &pid, &signal, &interval)) {
		goto out1;
	}

	if (!pid) {
		/* Clean checkpoint list */
		struct task_struct *p = NULL;
		printk(KERN_INFO "chpox: freeing checkpoint list...\n");
		down_write(&ckpt_lock);
		lh = ckpt_list.next;
		while(lh != &ckpt_list) {
			dt = list_entry(lh, struct ckpt_data, list);
			lh = lh->next;
			read_lock(&tasklist_lock);
			if ((p = find_task_by_pid(dt->pid))) {
				get_task_struct(p);
				read_unlock(&tasklist_lock);
				if ((current->euid ^ p->suid) && (current->euid ^ p->uid) &&
						(current->uid ^ p->suid) && (current->uid ^ p->uid) &&
						!capable(CAP_KILL)) {
					free_task_struct(p);
					goto no_remove;
				}
				spin_lock(&p->sigmask_lock);
				p->notifier = NULL;
				p->notifier_data = NULL;
				recalc_sigpending(p);
				spin_unlock(&p->sigmask_lock);
				free_task_struct(p);
				read_lock(&tasklist_lock);
			}
			read_unlock(&tasklist_lock);
			list_del(&dt->list);
			kfree(dt->file_name);
			kfree(dt);
			MOD_DEC_USE_COUNT;
no_remove:;
		}
		up_write(&ckpt_lock);
		goto out1;
	}

	if (signal > _NSIG) {
		ret = -EOVERFLOW;
		goto out1;
	}
      again:
	read_lock(&tasklist_lock);
	o = find_task_by_pid(pid);
	if (o)
		get_task_struct(o);
	read_unlock(&tasklist_lock);

	/* security check */
	if (o) {
		if ((current->euid ^ o->suid) && (current->euid ^ o->uid) &&
		    (current->uid ^ o->suid) && (current->uid ^ o->uid) &&
		    !capable(CAP_KILL)) {
			ret = -EPERM;
			goto out1;
		}
	}

	down_write(&ckpt_lock);
	if (interval == 0) {	/* remove from list */
		list_for_each(lh, &ckpt_list) {
			dt = list_entry(lh, struct ckpt_data, list);
			if (dt->pid == pid) {
				/*notifier is protected by &o->sigmask_lock */
				if (o) {
					spin_lock(&o->sigmask_lock);
					if (dt->count) {
						spin_unlock(&o->sigmask_lock);
						schedule();
						goto again;
					}
					o->notifier = NULL;
					o->notifier_data = NULL;
					recalc_sigpending(o);
					spin_unlock(&o->sigmask_lock);
				}
				list_del(lh);

				kfree(dt->file_name);
				kfree(dt);

				MOD_DEC_USE_COUNT;
				break;
			}
		}
		/* did not find dt entry */
		if (&ckpt_list == lh) {
			ret = -1;
		}
	}
	else {	/* add or modify list entry */
		if (o) {
			list_for_each(lh, &ckpt_list) {
				dt = list_entry(lh, struct ckpt_data, list);
				if (dt->pid == pid)
					break;
			}
			if (lh != &ckpt_list) {	/* already in list, modify entry */
				dt->interval = interval;
				dt->signal = signal;
			}
			else {	/* add new entry */

				if (create_chpox_entry(NULL, o, interval, signal) == -1)
					goto out;
			}
			/* if there were request try again later */

		}
		else {  /* task not found, delete it from list */
			list_for_each(lh, &ckpt_list) {
				dt = list_entry(lh, struct ckpt_data, list);
				if (dt->pid == pid) {
					MOD_DEC_USE_COUNT;
					list_del(lh);
					kfree(dt->file_name);
					kfree(dt);
					break;
				}
			}
		}
	}
      out:
	up_write(&ckpt_lock);
      out1:
	if (o)
		free_task_struct(o);
	MOD_DEC_USE_COUNT;
	return ret;
}

/*
*
*/
static int ckpt_read(char *page, char **start, off_t off, int count, int *eof,
	      void *data)
{
	struct ckpt_data *dt = NULL;
	struct list_head *lh = NULL;
	int len = 0, begin = 0;
	char *txt[] = {
		"C",		/*created */
		"S",		/* not used */
		"O",		/*OK*/
		"E",		/*error */
		NULL
	};
	MOD_INC_USE_COUNT;
	down_read(&ckpt_lock);
	list_for_each(lh, &ckpt_list) {
		dt = list_entry(lh, struct ckpt_data, list);
		len += sprintf(page + len,
			       "%d:%d:%d [%s|%d] -> %s [%ld.%ld]\n", dt->pid,
			       dt->signal, dt->interval, txt[dt->flags],
			       dt->nt, dt->file_name, dt->tv.tv_sec,
			       dt->tv.tv_usec);
		if (len + begin > off + count)
			break;
		if (len + begin < off) {
			begin += len;
			len = 0;
		}
	}
	if (lh == &ckpt_list)
		*eof = 1;
	up_read(&ckpt_lock);
	if (off >= len + begin) {
		MOD_DEC_USE_COUNT;
		return 0;
	}
	*start = page + (off - begin);
	MOD_DEC_USE_COUNT;
	return ((count < begin + len - off) ? count : begin + len - off);
}

/*

*/
/*--------------------------------------------------------------------
 *  modified binary format code from vmadump 
 *------------------------------------------------------------------*/
#ifdef HAVE_BINFMT_VMADUMP
static int load_chpox(struct linux_binprm *bprm, struct pt_regs *regs)
{
	int retval;
	struct vmadump_header *header;

	header = (struct vmadump_header *)bprm->buf;
	if (memcmp(header->magic, vmad_magic, sizeof(header->magic)) != 0 ||
	    header->fmt_vers != VMAD_FMT_VERS ||
	    header->major != ((LINUX_VERSION_CODE >> 16) & 0xFF) ||
	    header->minor != ((LINUX_VERSION_CODE >> 8) & 0xFF)
#if STRICT_VERSION_CHECK
	    || header->patch != (LINUX_VERSION_CODE & 0xFF)
#endif
		)
		return -ENOEXEC;

	if (!bprm->file->f_op || !bprm->file->f_op->mmap)
		return -ENOEXEC;

	retval = unchpox_proc(0, bprm->file, regs);
	if (retval == 0) {
		if (current->euid == current->uid
		    && current->egid == current->gid)
			current->mm->dumpable = 1;
		current->did_exec = 1;
		current->self_exec_id++;
		if (current->exec_domain && current->exec_domain->module)
			__MOD_DEC_USE_COUNT(current->exec_domain->module);
		if (current->binfmt && current->binfmt->module)
			__MOD_DEC_USE_COUNT(current->binfmt->module);
		current->exec_domain = 0;
		current->binfmt = 0;
	}
	return retval;
}

static struct linux_binfmt chpox_fmt = {
	0, THIS_MODULE, load_chpox, 0, 0, 0
};
extern struct linux_binfmt vmadump_fmt;

#endif
/*
*
*/

int init_module(void)
{
	printk(KERN_INFO "chpox: Olexander O. Sudakov "
	       "<csaa@mail.univ.kiev.ua>\n");
	dump_dir = proc_mkdir(DUMP_PROC_DIR, NULL);
	if (dump_dir) {
		dump_conf =
			create_proc_entry(DUMP_PROC_CONF,
					  S_IFREG | S_IRUGO | S_IWUSR,
					  dump_dir);
		dump_it =
			create_proc_entry(DUMP_PROC_REGISTER,
					  S_IFREG | S_IWUGO, dump_dir);
		dump_info =
			create_proc_entry(DUMP_PROC_INFO, S_IFREG | S_IRUGO,
					  dump_dir);
		if (!dump_conf || !dump_it || !dump_info) {
			remove_proc_entry(DUMP_PROC_CONF, dump_dir);
			remove_proc_entry(DUMP_PROC_REGISTER, dump_dir);
			remove_proc_entry(DUMP_PROC_INFO, dump_dir);
			remove_proc_entry(DUMP_PROC_DIR, NULL);
			printk(KERN_ERR "chpox: unable to create proc dir entries\n");
			return -1;
		}
		dump_it->write_proc = ckpt_write;
		dump_info->read_proc = ckpt_read;
	}
	else {
		printk(KERN_ERR "chpox: unable to create proc dir entry %s\n",
		       DUMP_PROC_DIR);
		return -1;
	}
#ifdef HAVE_BINFMT_VMADUMP
	unregister_binfmt(&vmadump_fmt);
	register_binfmt(&chpox_fmt);
#endif
	return 0;
}

/*
*
* */
void cleanup_module(void)
{
	unregister_binfmt(&chpox_fmt);
	register_binfmt(&vmadump_fmt);
	remove_proc_entry(DUMP_PROC_CONF, dump_dir);
	remove_proc_entry(DUMP_PROC_REGISTER, dump_dir);
	remove_proc_entry(DUMP_PROC_INFO, dump_dir);
	remove_proc_entry(DUMP_PROC_DIR, NULL);
}
